import os
from typing import AsyncGenerator

import pytest
from tests.utils import random_string

from scyllapy import Scylla


@pytest.fixture(scope="session")
def anyio_backend() -> str:
    """
    Anyio backend.

    Backend for anyio pytest plugin.
    :return: backend name.
    """
    return "asyncio"


@pytest.fixture(scope="session")
def scylla_url() -> str:
    return os.environ.get("SCYLLA_URL", "localhost:9042")


@pytest.fixture(scope="session")
async def keyspace(scylla_url: str) -> AsyncGenerator[str, None]:
    keyspace_name = random_string(5)
    scylla = Scylla(contact_points=[scylla_url])
    await scylla.startup()
    await scylla.execute(
        f"CREATE keyspace {keyspace_name} WITH replication = "
        "{'class': 'SimpleStrategy', 'replication_factor': 1}",
    )

    yield keyspace_name

    await scylla.execute(f"DROP KEYSPACE {keyspace_name}")


@pytest.fixture(scope="session")
async def scylla(scylla_url: str, keyspace: str) -> AsyncGenerator[Scylla, None]:
    scylla = Scylla(
        contact_points=[scylla_url],
        keyspace=keyspace,
    )
    await scylla.startup()

    yield scylla

    await scylla.shutdown()



import pytest
from tests.utils import random_string

from scyllapy import Scylla


@pytest.mark.anyio
async def test_results_len(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY)")
    for i in range(10):
        await scylla.execute(f"INSERT INTO {table_name}(id) VALUES (?)", [i])
    res = await scylla.execute(f"SELECT id FROM {table_name}")

    assert len(res) == 10



import pytest
from tests.utils import random_string

from scyllapy import Batch, Scylla


@pytest.mark.anyio
async def test_batches(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name}(id INT, PRIMARY KEY (id))")

    batch = Batch()
    num_queries = 10
    for _ in range(num_queries):
        batch.add_query(f"INSERT INTO {table_name}(id) VALUES (?)")

    await scylla.batch(batch, [[i] for i in range(num_queries)])

    res = await scylla.execute(f"SELECT id FROM {table_name}")
    assert set(res.scalars()) == set(range(num_queries))



import pytest
from tests.utils import random_string

from scyllapy import Scylla


@pytest.mark.anyio
async def test_prepared(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name}(id INT, PRIMARY KEY (id))")
    await scylla.execute(f"INSERT INTO {table_name}(id) VALUES (?)", [1])

    query = f"SELECT * FROM {table_name}"
    prepared = await scylla.prepare(query)
    res = await scylla.execute(query)
    prepared_res = await scylla.execute(prepared)

    assert res.all() == prepared_res.all()



import pytest
from tests.utils import random_string

from scyllapy import Scylla
from scyllapy.query_builder import Insert


@pytest.mark.anyio
async def test_insert_success(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await Insert(table_name).set("id", 1).set("name", "random").execute(scylla)
    result = await scylla.execute(f"SELECT * FROM {table_name}")
    assert result.all() == [{"id": 1, "name": "random"}]


@pytest.mark.anyio
async def test_insert_if_not_exists(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await Insert(table_name).set("id", 1).set("name", "random").execute(scylla)
    await Insert(table_name).set("id", 1).set(
        "name",
        "random2",
    ).if_not_exists().execute(scylla)
    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert res.all() == [{"id": 1, "name": "random"}]


@pytest.mark.anyio
async def test_insert_request_params(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await Insert(table_name).set("id", 1).set("name", "random").execute(scylla)
    res = (
        await Insert(table_name)
        .set("id", 1)
        .set("name", "random2")
        .request_params(
            tracing=True,
        )
        .execute(scylla)
    )
    assert res.trace_id



from ._internal.exceptions import (
    ScyllaPyBaseError,
    ScyllaPyBindingError,
    ScyllaPyDBError,
    ScyllaPyMappingError,
    ScyllaPyQueryBuiderErrror,
    ScyllaPySessionError,
)

__all__ = [
    "ScyllaPyBaseError",
    "ScyllaPyDBError",
    "ScyllaPySessionError",
    "ScyllaPyMappingError",
    "ScyllaPyQueryBuiderErrror",
    "ScyllaPyBindingError",
]



class BigInt:
    def __init__(self, val: int) -> None: ...

class SmallInt:
    def __init__(self, val: int) -> None: ...

class TinyInt:
    def __init__(self, val: int) -> None: ...

class Double:
    def __init__(self, val: float) -> None: ...

class Counter:
    def __init__(self, val: int) -> None: ...

class Unset:
    """
    Class for unsetting the variable.

    If you want to set NULL to a column,
    when performing INSERT statements,
    it's better to use Unset instead of setting
    NULL, because it may result in better performance.

    https://rust-driver.docs.scylladb.com/stable/queries/values.html#unset-values
    """

    def __init__(self) -> None: ...



from importlib.metadata import version

from ._internal import (
    Batch,
    BatchType,
    Consistency,
    InlineBatch,
    PreparedQuery,
    Query,
    QueryResult,
    Scylla,
    SerialConsistency,
)

__version__ = version("scyllapy")

__all__ = [
    "__version__",
    "Scylla",
    "Consistency",
    "Query",
    "SerialConsistency",
    "PreparedQuery",
    "Batch",
    "BatchType",
    "QueryResult",
    "extra_types",
    "InlineBatch",
    "query_builder",
]



class ScyllaPyBaseError(Exception):
    """Base scyllapy exception."""

class ScyllaPyBindingError(ScyllaPyBaseError):
    """
    Error that occurs during parameter binding.

    This error can be thrown if a parameter
    is not of the correct type and therefore cannot
    be bound.
    """

class ScyllaPyDBError(ScyllaPyBaseError):
    """
    Database related exception.

    This exception can be thrown when
    the database returns an error.
    """

class ScyllaPySessionError(ScyllaPyDBError):
    """
    Error related to database session.

    This exception can be thrown when
    session was not properly initialized,
    or if it was closed by some reason.
    """

class ScyllaPyMappingError(ScyllaPyBaseError):
    """
    Exception that occurs during mapping results back to python.

    It can be thrown if you request row fetching,
    but query didn't return any rows.

    Also it occurs if rows cannot be mapped to python types.
    """

class ScyllaPyQueryBuiderErrror(ScyllaPyBaseError):
    """
    Error that is thrown if Query cannot be built.

    When using query builder you can try to execute
    partialy built query that is guaranteed to
    have syntax errors. In order to avoid
    such situations we introduced another type of error,
    that is thrown before query is executed.
    """



import datetime
import ipaddress
import random
import uuid
from typing import Any, Callable

import pytest
from tests.utils import random_string

from scyllapy import Scylla


@pytest.mark.anyio
@pytest.mark.parametrize(
    ("type_name", "test_val"),
    [
        ("INT", 1),
        ("TEXT", "mytext"),
        ("VARCHAR", "text2"),
        ("ASCII", "randomtext"),
        ("BLOB", b"random_bytes"),
        ("BOOLEAN", True),
        ("BOOLEAN", False),
        ("DATE", datetime.date.today()),
        ("TIME", datetime.time(22, 30, 11, 403)),
        ("TIMEUUID", uuid.uuid1()),
        ("UUID", uuid.uuid1()),
        ("UUID", uuid.uuid3(uuid.uuid4(), "name")),
        ("UUID", uuid.uuid4()),
        ("UUID", uuid.uuid5(uuid.uuid4(), "name")),
        ("INET", ipaddress.ip_address("192.168.1.1")),
        ("INET", ipaddress.ip_address("2001:db8::8a2e:370:7334")),
    ],
)
async def test_bindings(
    scylla: Scylla,
    type_name: str,
    test_val: Any,
) -> None:
    table_name = random_string(4)
    await scylla.execute(
        f"CREATE TABLE {table_name} (id {type_name}, PRIMARY KEY (id))"
    )
    insert_query = f"INSERT INTO {table_name}(id) VALUES (?)"
    await scylla.execute(insert_query, [test_val])

    result = await scylla.execute(f"SELECT * FROM {table_name}")
    rows = result.all()
    assert len(rows) == 1
    assert rows[0] == {"id": test_val}


@pytest.mark.anyio
@pytest.mark.parametrize(
    ("type_name", "test_val", "cast_func"),
    [
        ("SET<TEXT>", ["one", "two"], set),
        ("SET<TEXT>", {"one", "two"}, set),
        ("SET<TEXT>", ("one", "two"), set),
        ("LIST<TEXT>", ("1", "2"), list),
        ("LIST<TEXT>", ["1", "2"], list),
        ("LIST<TEXT>", {"1", "2"}, list),
        ("MAP<TEXT, TEXT>", {"one": "two"}, dict),
    ],
)
async def test_collections(
    scylla: Scylla,
    type_name: str,
    test_val: Any,
    cast_func: Callable[[Any], Any],
) -> None:
    table_name = random_string(4)
    await scylla.execute(
        f"CREATE TABLE {table_name} (id INT, coll {type_name}, PRIMARY KEY (id))"
    )
    insert_query = f"INSERT INTO {table_name}(id, coll) VALUES (?, ?)"

    await scylla.execute(insert_query, [1, test_val])

    result = await scylla.execute(f"SELECT * FROM {table_name}")
    rows = result.all()
    assert len(rows) == 1
    assert rows[0] == {"id": 1, "coll": cast_func(test_val)}


@pytest.mark.anyio
async def test_named_parameters(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(
        f"CREATE TABLE {table_name} (id INT, name TEXT, age INT, PRIMARY KEY (id))"
    )
    to_insert = {
        "id": random.randint(0, 100),
        "name": random_string(5),
        "age": random.randint(0, 100),
    }
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name, age) VALUES (:id, :name, :age)",
        params=to_insert,
    )

    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert res.first() == to_insert


@pytest.mark.anyio
async def test_timestamps(scylla: Scylla) -> None:
    table_name = random_string(4)
    now = datetime.datetime.now()
    # We do replace this, because scylla ony has millisecond percision.
    now = now.replace(microsecond=now.microsecond - (now.microsecond % 1000))
    await scylla.execute(
        f"CREATE TABLE {table_name} (id INT, time TIMESTAMP, PRIMARY KEY (id))"
    )
    insert_query = f"INSERT INTO {table_name}(id, time) VALUES (?, ?)"

    await scylla.execute(insert_query, [1, now])

    res = await scylla.execute(f"SELECT time FROM {table_name}")
    assert res.scalar() == now


@pytest.mark.anyio
async def test_none_vals(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await scylla.execute(f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, None])
    results = await scylla.execute(f"SELECT * FROM {table_name}")
    assert results.first() == {"id": 1, "name": None}


@pytest.mark.anyio
async def test_cases(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (:Id, :NaMe)",
        {"Id": 1, "NaMe": 2},
    )



from typing import Any

from scyllapy._internal import (
    Consistency,
    InlineBatch,
    QueryResult,
    Scylla,
    SerialConsistency,
)

class Select:
    def __init__(self, table: str) -> None: ...
    def only(self, *columns: str) -> Select: ...
    def where(self, clause: str, params: list[Any] | None = None) -> Select: ...
    def group_by(self, group: str) -> Select: ...
    def order_by(self, order: str, desc: bool = False) -> Select: ...
    def per_partition_limit(self, per_partition_limit: int) -> Select: ...
    def limit(self, limit: int) -> Select: ...
    def allow_filtering(self) -> Select: ...
    def distinct(self) -> Select: ...
    def timeout(self, timeout: int | str) -> Select: ...
    def request_params(
        self,
        consistency: Consistency | None = None,
        serial_consistency: SerialConsistency | None = None,
        request_timeout: int | None = None,
        timestamp: int | None = None,
        is_idempotent: bool | None = None,
        tracing: bool | None = None,
    ) -> Select: ...
    def add_to_batch(self, batch: InlineBatch) -> None: ...
    async def execute(self, scylla: Scylla) -> QueryResult: ...

class Insert:
    def __init__(self, table: str) -> None: ...
    def if_not_exists(self) -> Insert: ...
    def set(self, name: str, value: Any) -> Insert: ...
    def timeout(self, timeout: int | str) -> Insert: ...
    def timestamp(self, timestamp: int) -> Insert: ...
    def ttl(self, ttl: int) -> Insert: ...
    def request_params(
        self,
        consistency: Consistency | None = None,
        serial_consistency: SerialConsistency | None = None,
        request_timeout: int | None = None,
        timestamp: int | None = None,
        is_idempotent: bool | None = None,
        tracing: bool | None = None,
    ) -> Insert: ...
    def add_to_batch(self, batch: InlineBatch) -> None: ...
    async def execute(self, scylla: Scylla) -> QueryResult: ...

class Delete:
    def __init__(self, table: str) -> None: ...
    def cols(self, *cols: str) -> Delete: ...
    def where(self, clause: str, values: list[Any] | None = None) -> Delete: ...
    def timeout(self, timeout: int | str) -> Delete: ...
    def timestamp(self, timestamp: int) -> Delete: ...
    def if_exists(self) -> Delete: ...
    def if_(self, clause: str, values: list[Any] | None = None) -> Delete: ...
    def request_params(
        self,
        consistency: Consistency | None = None,
        serial_consistency: SerialConsistency | None = None,
        request_timeout: int | None = None,
        timestamp: int | None = None,
        is_idempotent: bool | None = None,
        tracing: bool | None = None,
    ) -> Delete: ...
    def add_to_batch(self, batch: InlineBatch) -> None: ...
    async def execute(self, scylla: Scylla) -> QueryResult: ...

class Update:
    def __init__(self, table: str) -> None: ...
    def set(self, name: str, value: Any) -> Update: ...
    def inc(self, column: str, value: Any) -> Update: ...
    def dec(self, column: str, value: Any) -> Update: ...
    def where(self, clause: str, values: list[Any] | None = None) -> Update: ...
    def timeout(self, timeout: int | str) -> Update: ...
    def timestamp(self, timestamp: int) -> Update: ...
    def ttl(self, ttl: int) -> Update: ...
    def request_params(
        self,
        consistency: Consistency | None = None,
        serial_consistency: SerialConsistency | None = None,
        request_timeout: int | None = None,
        timestamp: int | None = None,
        is_idempotent: bool | None = None,
        tracing: bool | None = None,
    ) -> Update: ...
    def if_exists(self) -> Update: ...
    def if_(self, clause: str, values: list[Any] | None = None) -> Update: ...
    def add_to_batch(self, batch: InlineBatch) -> None: ...
    async def execute(self, scylla: Scylla) -> QueryResult: ...



import pytest
from tests.utils import random_string

from scyllapy import Scylla
from scyllapy.query_builder import Delete


@pytest.mark.anyio
async def test_success(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, "meme"]
    )
    await Delete(table_name).where("id = ?", [1]).execute(scylla)
    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert not res.all()


@pytest.mark.anyio
async def test_if_exists(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, "meme"]
    )
    await Delete(table_name).where("id = ?", [1]).if_exists().execute(scylla)
    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert not res.all()


@pytest.mark.anyio
async def test_custom_if(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, "meme"]
    )
    await Delete(table_name).where("id = ?", [1]).if_("name != ?", [None]).execute(
        scylla
    )
    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert not res.all()


@pytest.mark.anyio
async def test_custom_custom_if(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, "meme"]
    )
    await Delete(table_name).where("id = ?", [1]).if_("name != ?", [None]).execute(
        scylla
    )
    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert not res.all()



import uuid

import pytest
from tests.utils import random_string

from scyllapy import Scylla
from scyllapy.query_builder import Insert, Select


@pytest.mark.anyio
async def test_select_success(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, "meme"]
    )
    res = await Select(table_name).execute(scylla)
    assert res.all() == [{"id": 1, "name": "meme"}]


@pytest.mark.anyio
async def test_select_aliases(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    name = uuid.uuid4().hex
    await scylla.execute(f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, name])
    res = await Select(table_name).only("name as testname").execute(scylla)
    assert res.all() == [{"testname": name}]


@pytest.mark.anyio
async def test_select_simple_where(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    name = uuid.uuid4().hex
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, uuid.uuid4().hex]
    )
    await scylla.execute(f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [2, name])

    res = await Select(table_name).where("id = ?", [2]).execute(scylla)
    assert res.all() == [{"id": 2, "name": name}]


@pytest.mark.anyio
async def test_select_multiple_filters(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(
        f"CREATE TABLE {table_name} (id INT, name TEXT, PRIMARY KEY (id, name))"
    )
    name = uuid.uuid4().hex
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, uuid.uuid4().hex]
    )
    await scylla.execute(f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [2, name])

    res = (
        await Select(table_name)
        .where("id = ?", [2])
        .where("name = ?", [name])
        .execute(scylla)
    )
    assert res.all() == [{"id": 2, "name": name}]


@pytest.mark.anyio
async def test_allow_filtering(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    name = uuid.uuid4().hex
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, uuid.uuid4().hex]
    )
    await scylla.execute(f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [2, name])

    res = (
        await Select(table_name)
        .where("id = ?", [2])
        .where("name = ?", [name])
        .allow_filtering()
        .execute(scylla)
    )
    assert res.all() == [{"id": 2, "name": name}]


@pytest.mark.anyio
async def test_limit(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    for i in range(10):
        await scylla.execute(
            f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [i, uuid.uuid4().hex]
        )
    res = await Select(table_name).limit(3).execute(scylla)
    assert len(res.all()) == 3


@pytest.mark.anyio
async def test_order_by(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(
        f"CREATE TABLE {table_name} (id INT, iid INT, PRIMARY KEY(id, iid))"
    )
    for i in range(10):
        await scylla.execute(
            f"INSERT INTO {table_name}(id, iid) VALUES (?, ?)",
            [0, i],
        )
    res = (
        await Select(table_name)
        .only("iid")
        .where("id = ?", [0])
        .order_by("iid")
        .execute(scylla)
    )
    ids = [row["iid"] for row in res.all()]
    assert ids == list(range(10))



import pytest
from tests.utils import random_string

from scyllapy import Scylla
from scyllapy.query_builder import Update


@pytest.mark.anyio
async def test_success(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, "meme"]
    )
    await Update(table_name).set("name", "meme2").where("id = ?", [1]).execute(scylla)
    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert res.all() == [{"id": 1, "name": "meme2"}]


@pytest.mark.anyio
async def test_ifs(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await scylla.execute(
        f"INSERT INTO {table_name}(id, name) VALUES (?, ?)", [1, "meme"]
    )
    await Update(table_name).set("name", "meme2").if_("name = ?", ["meme"]).where(
        "id = ?", [1]
    ).execute(scylla)
    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert res.all() == [{"id": 1, "name": "meme2"}]


@pytest.mark.anyio
async def test_if_exists(scylla: Scylla) -> None:
    table_name = random_string(4)
    await scylla.execute(f"CREATE TABLE {table_name} (id INT PRIMARY KEY, name TEXT)")
    await Update(table_name).set("name", "meme2").if_exists().where(
        "id = ?", [1]
    ).execute(scylla)
    res = await scylla.execute(f"SELECT * FROM {table_name}")
    assert res.all() == []



from typing import Any, Callable, Iterable, Literal, Optional, TypeVar, overload

T = TypeVar("T")

class Scylla:
    """
    Scylla class.

    This class represents scylla cluster.
    And has internal connection pool.

    Everything that can beconfigured, shown below.
    """

    def __init__(
        self,
        contact_points: list[str],
        username: str | None = None,
        password: str | None = None,
        keyspace: str | None = None,
        ssl_cert: str | None = None,
        conn_timeout: int | None = None,
        write_coalescing: bool | None = None,
        pool_size_per_host: int | None = None,
        pool_size_per_shard: int | None = None,
        keepalive_interval: int | None = None,
        keepalive_timeout: int | None = None,
        tcp_keepalive_interval: int | None = None,
        tcp_nodelay: bool | None = None,
        disallow_shard_aware_port: bool | None = None,
    ) -> None:
        """
        Configure cluster for later use.

        :param contact_points: List of known nodes. (Hosts and ports).
            ["192.168.1.1:9042", "my_keyspace.node:9042"]
        :param username: Plain text auth username.
        :param password: Plain text auth password.
        :param ssl_cert: Certficiate string to use
            for connection. AWS requires it.
        :param conn_timeout: Timeout in seconds.
        :param write_coalescing:
            If true, the driver will inject a small delay before flushing data
            to the socket - by rescheduling the task that writes data to the socket.
            This gives the task an opportunity to collect more write requests
            and write them in a single syscall, increasing the efficiency.
        :param pool_size_per_host: how many connections should be established
            to the node.
        :param pool_size_per_host: how many connections should be established
            to each shard of the node.
        :param keepalive_interval: How ofter to send keepalive messages,
            when connection is idling. In seconds.
        :param keepalive_timeout: sets keepalive timeout.
        :param tcp_keepalive_interval: Sets TCP keepalive interval.
        :param tcp_nodelay: sets TCP nodelay flag.
        :param disallow_shard_aware_port: If true, prevents the driver from connecting to the shard-aware port,
            even if the node supports it.
        """
    async def startup(self) -> None:
        """Initialize the custer."""
    async def shutdown(self) -> None:
        """Shutdown the cluster."""
    async def prepare(self, query: str | Query) -> PreparedQuery: ...
    async def execute(
        self,
        query: str | Query | PreparedQuery,
        params: Optional[Iterable[Any] | dict[str, Any]] = None,
    ) -> QueryResult:
        """
        Execute a query.

        This function takes a query string,
        and list of parameters.

        Parameters in query can be specified as ? signs.

        await scylla.execute("SELECT * FROM table WHERE id = ?", [11])

        Or you can use named parameters and pass dict to execute. Like this:

        await scylla.execute("SELECT * FROM table WHERE id = :id", {"id": 11})

        :param query: query to use.
        :param params: list of query parameters.
        :param as_class: DTO class to use for parsing rows (Can be pydantic model or dataclass).
        """
    async def batch(
        self,
        batch: Batch | InlineBatch,
        params: Optional[Iterable[Iterable[Any] | dict[str, Any]]] = None,
    ) -> QueryResult:
        """
        Execute a batch statement.

        Batch statements are useful for grouping multiple queries
        together and executing them in one query.

        Each element of a list associated

        It may speed up you application.
        """
    async def use_keyspace(self, keyspace: str) -> None:
        """Change current keyspace for all connections."""
    async def get_keyspace(self) -> str | None:
        """Get current keyspace."""

class QueryResult:
    trace_id: str | None

    @overload
    def all(self, as_class: Literal[None] = None) -> list[dict[str, Any]]: ...
    @overload
    def all(self, as_class: Callable[..., T] | None = None) -> list[T]: ...
    @overload
    def first(self, as_class: Literal[None] = None) -> dict[str, Any] | None: ...
    @overload
    def first(self, as_class: Callable[..., T] | None = None) -> T | None: ...
    def scalars(self) -> list[Any]: ...
    def scalar(self) -> Any | None: ...
    def __len__(self) -> int: ...

class Query:
    """
    Query class.

    It's used for fine-tuning specific queries.
    If you don't need a specific consistency, or
    any other parameter, you can pass a string instead.
    """

    query: str
    consistency: Consistency | None
    serial_consistency: SerialConsistency | None
    request_timeout: int | None
    is_idempotent: bool | None
    tracing: bool | None

    def __init__(
        self,
        query: str,
        consistency: Consistency | None = None,
        serial_consistency: SerialConsistency | None = None,
        request_timeout: int | None = None,
        timestamp: int | None = None,
        is_idempotent: bool | None = None,
        tracing: bool | None = None,
    ) -> None: ...
    def with_consistency(self, consistency: Consistency | None) -> Query: ...
    def with_serial_consistency(
        self,
        serial_consistency: SerialConsistency | None,
    ) -> Query: ...
    def with_request_timeout(self, request_timeout: int | None) -> Query: ...
    def with_timestamp(self, timestamp: int | None) -> Query: ...
    def with_is_idempotent(self, is_idempotent: bool | None) -> Query: ...
    def with_tracing(self, tracing: bool | None) -> Query: ...

class BatchType:
    """Possible BatchTypes."""

    COUNTER: BatchType
    LOGGED: BatchType
    UNLOGGED: BatchType

class Batch:
    """Class for batching queries together."""

    def __init__(
        self,
        batch_type: BatchType = BatchType.UNLOGGED,
        consistency: Consistency | None = None,
        serial_consistency: SerialConsistency | None = None,
        request_timeout: int | None = None,
        timestamp: int | None = None,
        is_idempotent: bool | None = None,
        tracing: bool | None = None,
    ) -> None: ...
    def add_query(self, query: Query | PreparedQuery | str) -> None: ...

class InlineBatch:
    def __init__(
        self,
        batch_type: BatchType = BatchType.UNLOGGED,
        consistency: Consistency | None = None,
        serial_consistency: SerialConsistency | None = None,
        request_timeout: int | None = None,
        timestamp: int | None = None,
        is_idempotent: bool | None = None,
        tracing: bool | None = None,
    ) -> None: ...
    def add_query(
        self,
        query: Query | PreparedQuery | str,
        values: list[Any] | None = None,
    ) -> None: ...

class Consistency:
    """Consistency for query."""

    ANY: "Consistency"
    ONE: "Consistency"
    TWO: "Consistency"
    THREE: "Consistency"
    QUORUM: "Consistency"
    ALL: "Consistency"
    LOCAL_QUORUM: "Consistency"
    EACH_QUORUM: "Consistency"
    LOCAL_ONE: "Consistency"

class SerialConsistency:
    """Serial consistency for query."""

    SERIAL: SerialConsistency
    LOCAL_SERIAL: SerialConsistency

class PreparedQuery:
    """Class that represents prepared statement."""
